#NoTrayIcon
#Region ;**** Directives created by AutoIt3Wrapper_GUI ****
#AutoIt3Wrapper_Icon=MicSwitcher.ico
#AutoIt3Wrapper_Compression=3
#AutoIt3Wrapper_Compile_Both=y
#AutoIt3Wrapper_UseX64=y
#AutoIt3Wrapper_Res_Comment=beta software
#AutoIt3Wrapper_Res_Description=Mic. Switcher
#AutoIt3Wrapper_Res_Fileversion=0.2024.1.6
#AutoIt3Wrapper_Res_ProductName=MicSwitcher.exe
#AutoIt3Wrapper_Res_ProductVersion=0.0
#AutoIt3Wrapper_Res_CompanyName=none
#AutoIt3Wrapper_Res_LegalCopyright=none
#AutoIt3Wrapper_Res_LegalTradeMarks=none
#AutoIt3Wrapper_Res_Language=1000
#AutoIt3Wrapper_Res_Field=ProductName|%scriptfile%.exe
#AutoIt3Wrapper_Res_Field=Build Date|%date% %time%
#AutoIt3Wrapper_Res_Icon_Add=MicSwitcher.speakers.ico,201
#AutoIt3Wrapper_Res_Icon_Add=MicSwitcher.Headset.ico,202
#AutoIt3Wrapper_AU3Check_Parameters=-q -d -w 1 -w 2 -w 3 -w 4 -w 5 -w 6 -w 7
#AutoIt3Wrapper_Run_Au3Stripper=y
#Au3Stripper_Parameters=/MO
#EndRegion ;**** Directives created by AutoIt3Wrapper_GUI ****
;~ #RequireAdmin

#cs

	Do run/compile with v3.3.14.5, "OnEvent" is too slow in newer versions.
	( get it at https://www.autoitscript.com/autoit3/files/archive/autoit/autoit-v3.3.14.5.zip )

	https://www.autoitscript.com/forum/files/file/476-setmicforplayback/
	https://www.autoitscript.com/forum/topic/196727-set-microphone-for-playback-device/

	What's New in Version 2.2024.1.6
	Added: Move "Volume Mixer" to righ-bottom corner option ( mostly for Win11 )
	Added: Expand the "Volume Mixer" window by N pixels. ( set in the ini file to change or disable )
	Added: "CTRL + click tray" or "middle mouse button (three-button mouse)", will load "App volume and device preferences".
	Fixed: running as script now works without the need to compile first.

#ce

Opt("MustDeclareVars", 1)
;~ #include <OnDebugMsgBox.au3> ; https://www.autoitscript.com/forum/files/file/500-ondebugmsgbox/
#include <WinAPIGdi.au3> ; _WinAPI_GetMonitorInfo()
#include <WinAPIMisc.au3> ; _WinAPI_GetMousePos()
#include <File.au3>
#include <Debug.au3>
#include <GuiMenu.au3>
#include <TrayConstants.au3> ; Required For the $TRAY_EVENT_PRIMARYDOUBLE, $TRAY_EVENT_SECONDARYUP And $TRAY_ICONSTATE_SHOW constants.
#include <GUIConstantsEx.au3>
#include <WindowsConstants.au3>
#include <Constants.au3>
#include <GUIConstants.au3>
#include <APIRegConstants.au3>
#include <MsgBoxConstants.au3>
#include <WinAPIError.au3>
#include <WinAPIHObj.au3>
#include <WinAPIProc.au3>
#include <WinAPIReg.au3>
#include <WinAPICom.au3>
#include <WinAPIConv.au3>
#include <WinAPISys.au3>
#include <Date.au3>
#include <SQLite.au3> ; _SQLite_Display2DResult()
#include <Misc.au3> ; _IsPressed()

Global $g_ExeName = StringTrimRight(@ScriptName, 4)
If Not @Compiled And Not FileGetSize(@ScriptDir & "\" & $g_ExeName & " - Shortcut.lnk") Then
	FileCreateShortcut(@AutoItExe,   @ScriptDir & "\" & $g_ExeName & " - Shortcut", @ScriptDir, ' "' & @ScriptFullPath & '" ', "alt. tray sound menu", StringTrimRight(@ScriptFullPath, 4) & ".ico")
EndIf

Global $g_IfMiddleClick_RunAppVolume = TimerInit() ; for Win11 WTFs
Global $g_EVENT_MOUSEOVER_aMousePos[2] = [10,10], $g_EVENT_MOUSEOVER_hTimer = TimerInit()
Global $hGui_GuiContextMenu = 0, $hGui_Main = 0
Global $g_EventListener_tEvents = 0, $g_EventListener_hKey = 0
gTimer() ; init. global timer init
Global $g_ConsoleLogToFile = IniGet("LogToFile", 1)
Global $g_ResizeVolMix = IniGet("ResizeVolMix", 300, Default, " ; in pixels. Set to zero to disable resizing.") ;
Global $g_MoveToBottomRightVolMix = IniGet("MoveToBottomRightVolMix", 1, Default, ' ; for fixing Win11 GUI position, forcing users to use their -much disliked by me- "new and improved" interface')
Global $g_OnClickVolMix = IniGet("OnClickVolMix", 1)
Global $g_AutoCloseVolMix = IniGet("AutoCloseVolMix", 1)
Global $g_AutoCloseVolMix_Timer = TimerInit()
Global $g_UserMicSel = StringSplit(IniGet("UserMicSel") & "||", "|", 2)
Global $g_MyTrayMenuData
Global $sErrorHandler = "", $iErrorHandler = 0, $oErrorHandler = ObjEvent("AutoIt.Error", "_ErrFunc")
Func _ErrFunc($oError)
	Local $s = "COM Error, Line(" & $oError.scriptline & ") : Number 0x" & Hex($oError.number, 8) & " - " & $oError.windescription & @CRLF
	$sErrorHandler &= $s
	$iErrorHandler += 1
EndFunc   ;==>_ErrFunc

#Region IMMDevice

Global Enum $eConsole, $eMultimedia, $eCommunications, $ERole_enum_count
Global Const $sCLSID_CPolicyConfigClient = "{870af99c-171d-4f9e-af0d-e63df40c2bc9}"
Global Const $sIID_IPolicyConfig = "{f8679f50-850a-41cf-9c72-430f290290c8}"

Global Const $tagIPolicyConfig = "GetMixFormat hresult(wstr;ptr*);" & _
		"GetDeviceFormat hresult(wstr;int;ptr*);" & _
		"ResetDeviceFormat hresult(wstr);" & _
		"SetDeviceFormat hresult(wstr;ptr;ptr);" & _
		"GetProcessingPeriod hresult(wstr;int;int64*;int64*);" & _
		"SetProcessingPeriod hresult(wstr;int64*);" & _
		"GetShareMode hresult(wstr;ptr);" & _
		"SetShareMode hresult(wstr;ptr);" & _
		"GetPropertyValue hresult(wstr;struct;variant*);" & _
		"SetPropertyValue hresult(wstr;struct;variant*);" & _
		"SetDefaultEndpoint hresult(wstr;int);" & _
		"SetEndpointVisibility hresult(wstr;int);"



Func _SwitchToDevice($sId) ; run separatedly ( another PID, let AutoIt "Garbage Clean" )
	Local Const $S_OK = 0
;~ 	Local Const $S_FALSE = 1
	Local $hResult, $oPolicyConfig = ObjCreateInterface($sCLSID_CPolicyConfigClient, $sIID_IPolicyConfig, $tagIPolicyConfig)
	$hResult = $oPolicyConfig.SetDefaultEndpoint($sId, $eConsole)
	If $hResult = $S_OK Then $hResult = $oPolicyConfig.SetDefaultEndpoint($sId, $eCommunications)
	$oPolicyConfig = ""
	Return $hResult = $S_OK
EndFunc   ;==>_SwitchToDevice


Global Enum $eRender, $eCapture, $eAll, $EDataFlow_enum_count
Global Const $sCLSID_MMDeviceEnumerator = "{BCDE0395-E52F-467C-8E3D-C4579291692E}"
Global Const $sIID_IMMDeviceEnumerator = "{A95664D2-9614-4F35-A746-DE8DB63617E6}"

Global Const $tagIMMDeviceEnumerator = "EnumAudioEndpoints hresult(dword;dword;ptr*);" & _
		"GetDefaultAudioEndpoint hresult(dword;dword;ptr*);" & _
		"GetDevice hresult(wstr;ptr*);" & _
		"RegisterEndpointNotificationCallback hresult(ptr);" & _
		"UnregisterEndpointNotificationCallback hresult(ptr);"

Global Const $STGM_READ = 0
Global Const $DEVICE_STATE_ACTIVE = 0x00000001
Global Const $IID_IMMDeviceCollection = "{0BD7A1BE-7A1A-44DB-8397-CC5392387B5E}"
Global Const $tagIMMDeviceCollection = "GetCount hresult(uint*);" & _
		"Item hresult(uint;ptr*)"

Global Const $sPKEY_Device_FriendlyName = "{a45c254e-df1c-4efd-8020-67d146a850e0} 14"
Global Const $IID_IMMDevice = "{D666063F-1587-4E43-81F1-B948E807363F}"
Global Const $tagIMMDevice = "Activate hresult(ptr;dword;variant*;ptr*);" & _
		"OpenPropertyStore hresult(dword;ptr*);" & _
		"GetId hresult(ptr*);" & _
		"GetState hresult(dword*);"

Global Const $IID_IPropertyStore = "{886d8eeb-8cf2-4446-8d02-cdba1dbdcf99}"
Global Const $tagIPropertyStore = "GetCount hresult(dword*);" & _
		"GetAt hresult(dword;ptr*);" & _
		"GetValue hresult(ptr;variant*);" & _
		"SetValue hresult(ptr;variant);" & _
		"Commit hresult();"

Func _EnumerateAudioDevices()
	Local $idDefRender = _GetDefaultEndpoint($eRender)
	Local $idDefCapture = _GetDefaultEndpoint($eCapture)

;~ Registry Key for Audio Devices
	Local $sRegAudio = @OSArch = "x64" ? "HKLM64\SOFTWARE\Microsoft\Windows\CurrentVersion\MMDevices\Audio" : "HKLM\SOFTWARE\Microsoft\Windows\CurrentVersion\MMDevices\Audio"

;~ Create MMDeviceEnumerator object
	Local $oEnumerator = ObjCreateInterface($sCLSID_MMDeviceEnumerator, $sIID_IMMDeviceEnumerator, $tagIMMDeviceEnumerator)

;~ Get collection out of it
	Local $pCollection
	$oEnumerator.EnumAudioEndpoints($eAll, $DEVICE_STATE_ACTIVE, $pCollection)
;~ Turn it into object
	Local $oCollection = ObjCreateInterface($pCollection, $IID_IMMDeviceCollection, $tagIMMDeviceCollection)

;~ Check the number of devices in collection
	Local $iCount
	$oCollection.GetCount($iCount)

;~ Array for out
	Local $aArray[$iCount][6]
	Local $pEndpoint, $oEndpoint
	Local $pID, $sId, $pProps, $oProps
	Local $sName, $aGuid
	Local $tPKEY_Device_FriendlyName = _WinAPI_PKEYFromString($sPKEY_Device_FriendlyName)

;~ Actual enumeration
	For $i = 0 To $iCount - 1
;~ Item at "i" index is audio endpoint device
		$oCollection.Item($i, $pEndpoint)
		$oEndpoint = ObjCreateInterface($pEndpoint, $IID_IMMDevice, $tagIMMDevice)

;~ Collect ID
		$oEndpoint.GetId($pID)
		$sId = DllStructGetData(DllStructCreate("wchar ID[" & _WinAPI_PtrStringLenW($pID) & "]", $pID), "ID")
		_WinAPI_CoTaskMemFree($pID)
		$aArray[$i][2] = $sId
		$aArray[$i][4] = 0
		If $sId = $idDefRender Or $sId = $idDefCapture Then $aArray[$i][4] = 1

;~ Detect Instance ID is Microphone(capture) or Playback(render)
		$aGuid = StringSplit($sId, ".")
		$aArray[$i][1] = $aGuid[$aGuid[0]]
		$aArray[$i][0] = RegRead($sRegAudio & "\Capture\" & $aGuid[$aGuid[0]], "DeviceState") >= 1 ? "Microphone" : ""
		If $aArray[$i][0] = "" Then $aArray[$i][0] = RegRead($sRegAudio & "\Render\" & $aGuid[$aGuid[0]], "DeviceState") >= 1 ? "Playback" : ""
		If $aArray[$i][0] = "" Then $aArray[$i][0] = RegRead($sRegAudio & "\RemoteCapture\" & $aGuid[$aGuid[0]], "Protocol") >= 1 ? "Microphone" : ""
		If $aArray[$i][0] = "" Then $aArray[$i][0] = RegRead($sRegAudio & "\RemoteRender\" & $aGuid[$aGuid[0]], "Protocol") >= 1 ? "Playback" : ""

;~ PropertyStore stores properties, lol
		$oEndpoint.OpenPropertyStore($STGM_READ, $pProps)
		$oProps = ObjCreateInterface($pProps, $IID_IPropertyStore, $tagIPropertyStore)

;~ Collect FriendlyName property
		$sName = 0
		$oProps.GetValue(DllStructGetPtr($tPKEY_Device_FriendlyName), $sName)
		$aArray[$i][3] = $sName
		$aArray[$i][5] = StringForArraySort($sName)
	Next
	$oProps = ""
	$oEndpoint = ""
	$oCollection = ""
	_ArraySort($aArray, 0, 0, 0, 5)
	ReDim $aArray[$i + 1][5]
	Return $aArray
EndFunc   ;==>_EnumerateAudioDevices

Func StringForArraySort($s)
	Local $n, $v, $a = StringSplit($s, " (-)", 0)
	For $n = 1 To $a[0]
		$v &= StringLeft($a[$n], 1)
	Next
	Return $v
EndFunc   ;==>StringForArraySort

Func _GetDefaultEndpoint($iType = $eRender)
	Local $pEndpoint, $oEndpoint
	Local $pID, $sId
;~ 	Local $tPKEY_Device_FriendlyName = _WinAPI_PKEYFromString($sPKEY_Device_FriendlyName)
	Local $oEnumerator = ObjCreateInterface($sCLSID_MMDeviceEnumerator, $sIID_IMMDeviceEnumerator, $tagIMMDeviceEnumerator)
	$oEnumerator.GetDefaultAudioEndpoint($iType, $DEVICE_STATE_ACTIVE, $pEndpoint)
	If $pEndpoint = 0 Then Return SetError(1, 0, "GetDefaultAudioEndpoint failed")
	$oEndpoint = ObjCreateInterface($pEndpoint, $IID_IMMDevice, $tagIMMDevice)
	$oEndpoint.GetId($pID)
	$sId = DllStructGetData(DllStructCreate("wchar ID[" & _WinAPI_PtrStringLenW($pID) & "]", $pID), "ID")
	_WinAPI_CoTaskMemFree($pID)
	Return $sId
EndFunc   ;==>_GetDefaultEndpoint

Func _WinAPI_PtrStringLenW($pString)
	Local $aCall = DllCall("kernel32.dll", "dword", "lstrlenW", "ptr", $pString)
	If @error Then Return SetError(1, 0, 0)
	Return $aCall[0]
EndFunc   ;==>_WinAPI_PtrStringLenW

Func _WinAPI_PKEYFromString($sPKEY, $pID = Default)
	Local $tPKEY = DllStructCreate("byte GUID[16]; dword PID;")
	DllCall("propsys.dll", "long", "PSPropertyKeyFromString", "wstr", $sPKEY, "ptr", DllStructGetPtr($tPKEY))
	If @error Then ; alternative for unsupported systems
		Local $tGUID = _WinAPI_GUIDFromString($sPKEY)
		If Not @error Then
			DllStructSetData($tPKEY, "GUID", DllStructGetData(DllStructCreate("byte Data[16]", DllStructGetPtr($tGUID)), 1))
		EndIf
		DllStructSetData($tPKEY, "PID", Number(StringRegExpReplace($sPKEY, ".*?}", "")))
	EndIf
	If $pID <> Default Then DllStructSetData($tPKEY, "PID", $pID)
	Return $tPKEY
EndFunc   ;==>_WinAPI_PKEYFromString

#EndRegion IMMDevice

CmdLine_Eval()
Func CmdLine_Eval()
	Local $n, $i
	For $n = 1 To $CmdLine[0]
		If StringInStr($CmdLine[$n], Chr(1) & "Starting:" & Chr(1), 1) Then
			WinClose($g_ExeName & "_MemLeakShowArray")
			AutoItWinSetTitle($g_ExeName & "_MemLeakShowArray")
			Local $a = __ArrayFromString($CmdLine[$n], Chr(1), Chr(2))
			$a[0][2] = $a[0][0]
			$a[0][0] = "RunTime"
			$a[0][1] = ""
			$a[0][3] = ""
			For $n = 0 To UBound($a) - 1
				$a[$n][0] &= ": "
			Next
			_DebugArrayDisplay($a, $g_ExeName & " :  Resources usage  (for debug)", Default, Default, @TAB & "|", " |Initial|Now|Diff. (multp.)")
			Exit 0
		EndIf
		If StringInStr($CmdLine[$n], "}.{", 1) Then
			$i = _SwitchToDevice($CmdLine[$n])
			If $i Then
				Exit 3
			Else
				Exit 4
			EndIf
		EndIf
		If StringInStr($CmdLine[$n], "/DoEnumerateAudioDevices") Then
			WinClose($g_ExeName & "_DoEnumerateAudioDevices")
			AutoItWinSetTitle($g_ExeName & "_DoEnumerateAudioDevices")
			_DoEnumerateAudioDevices()
			Exit 0
		EndIf
		If StringInStr($CmdLine[$n], "/ForceCrash") Then $CmdLine[100] = "this is a crash test  =)"
		If StringInStr($CmdLine[$n], "/About") Then
			About_Box()
			Exit 0
		EndIf
	Next
	If WinExists($g_ExeName & "_Main") Then
		WinClose($g_ExeName & "_Main")
		If Not WinWaitClose($g_ExeName & "_Main", "", 5) Then Exit 14
	EndIf
	main()
EndFunc   ;==>CmdLine_Eval

#Region tray

Func ShowResourcesUsage()
	Local $s, $a = MemLeakUsage() ; ..you could call this on the fly and get an array
	_Array_Rotate2D($a)
	$s = _ArrayToString($a, Chr(1), -1, -1, Chr(2))
	ShellExecute(@AutoItExe, ( @Compiled ? "" : '"' & @ScriptFullPath & '" ' ) & '"' & $s & '"')
EndFunc   ;==>ShowResourcesUsage

Func MyTrayMenuReqData()
	Local Static $t = 1
	$g_EVENT_MOUSEOVER_hTimer = TimerInit()
	$g_EVENT_MOUSEOVER_aMousePos = MouseGetPos()
	If @error Then Dim $g_EVENT_MOUSEOVER_aMousePos[2] = [10,10]

	If TimerDiff($t) > 1000 Then $g_MyTrayMenuData = _EnumerateAudioDevices()
	$t = TimerInit()
EndFunc   ;==>MyTrayMenuReqData

Func DesktopWorkArea($wh)
	Local Static $aWorkArea[2] = [-2,-2]
	If $aWorkArea[0] = -2 Then
		Local $tPos = _WinAPI_GetMousePos()
		Local $hMonitor = _WinAPI_MonitorFromPoint($tPos)
		Local $aData = _WinAPI_GetMonitorInfo($hMonitor)
		If IsArray($aData) Then
			$aWorkArea[0] = DllStructGetData($aData[1], 3)
			$aWorkArea[1] = DllStructGetData($aData[1], 4)
		Else
			$aWorkArea[0] = @DesktopWidth
			$aWorkArea[1] = @DesktopHeight
		EndIf
	EndIf
	Return $aWorkArea[$wh] ; can be 0=W or 1=H only.
EndFunc

Func MyTrayMenuVolMix()
	If TimerDiff($g_IfMiddleClick_RunAppVolume) < 1000 Then Return
	Local Static $exeSndVol = "", $sSndVolTitle = "[TITLE:Volume Mixer - ;CLASS:#32770;]"
	Local $iMouseGetPos_Error, $aPos = MouseGetPos()
	$iMouseGetPos_Error = @error

	If _IsPressed("11") Then ; 11 CTRL key ; App volume and device preferences
		ShellExecute("ms-settings:apps-volume")
		Return
	EndIf

	If $exeSndVol = "" Then
		If Int(FileGetVersion("SndVol.exe")) Then $exeSndVol = "SndVol.exe"
	EndIf
	If $exeSndVol = "" Then
		If Int(FileGetVersion("SndVol32.exe")) Then $exeSndVol = "SndVol32.exe"
	EndIf
	If $exeSndVol = "" Then Return
	If TimerDiff($g_AutoCloseVolMix_Timer) < 300 Then Return ; added v0.2019.1.17 ; remove annoyance, to have to click twice when $g_AutoCloseVolMix = 1
	If WinExists($sSndVolTitle) Then Return WinClose($sSndVolTitle)
	If $iMouseGetPos_Error Then Return
	ShellExecute($exeSndVol, "-m " & ($aPos[1] * 65536) + $aPos[0])
	If WinWaitActive($sSndVolTitle, "", 15) Then ; added v0.2021.4.14 ; "Volume Mixer" would not close when loosing focus otherwise
		If $g_ResizeVolMix Then
			Local $aWinPos = WinGetPos($sSndVolTitle)
			If Not @error Then
				If $g_MoveToBottomRightVolMix Then
					WinMove($sSndVolTitle, "", DesktopWorkArea(0) - $aWinPos[2] - $g_ResizeVolMix, DesktopWorkArea(1) - $aWinPos[3], $aWinPos[2] + $g_ResizeVolMix, $aWinPos[3]) ; 2024.01.05
				Else
					If $aWinPos[0] + $aWinPos[1] = 0 Then ; Win11 crap. Braking their own code.
						WinMove($sSndVolTitle, "", $aWinPos[0], $aWinPos[1], $aWinPos[2] + $g_ResizeVolMix, $aWinPos[3])
					Else
						WinMove($sSndVolTitle, "", $aWinPos[0] - $g_ResizeVolMix, $aWinPos[1], $aWinPos[2] + $g_ResizeVolMix, $aWinPos[3]) ; 2024.01.05
					EndIf
				EndIf
				Sleep(200)
			EndIf
		EndIf
		If $g_AutoCloseVolMix Then AdlibRegister("AutoCloseVolMix", 50)
	EndIf
EndFunc   ;==>MyTrayMenuVolMix

Func AutoCloseVolMix()
	If Not WinActive("[TITLE:Volume Mixer - ;CLASS:#32770;]") Then
		$g_AutoCloseVolMix_Timer = TimerInit()
		AdlibUnRegister("AutoCloseVolMix")
		WinClose("[TITLE:Volume Mixer - ;CLASS:#32770;]")
	EndIf
EndFunc   ;==>AutoCloseVolMix

Func MyTrayMenu()
	Local $m = MouseGetPos();, $t = TimerInit()
;~ 		sndvol.exe -m 26214900
;~ 		This example creates window at x=500 y=400
;~ 		coordinates = y * 65536 + x, where x and y - signed integers
	If _IsPressed("11") And _IsPressed("10") Then ; 10 SHIFT key & 11 CTRL key
		ShowResourcesUsage()
		Return
	EndIf
	Local Static $exeMsdt = ""
	If $exeMsdt = "" Then
		If Int(FileGetVersion("msdt.exe")) Then $exeMsdt = "msdt.exe"
	EndIf

	Local Static $exeSndVol = ""
	If $exeSndVol = "" Then
		If Int(FileGetVersion("SndVol.exe")) Then $exeSndVol = "SndVol.exe"
	EndIf
	If $exeSndVol = "" Then
		If Int(FileGetVersion("SndVol32.exe")) Then $exeSndVol = "SndVol32.exe"
	EndIf

	If WinExists("[TITLE:Volume Mixer - ;CLASS:#32770;]") Then WinClose("[TITLE:Volume Mixer - ;CLASS:#32770;]")

	Local Enum $e_idCancel = 1000, $e_idSave, $e_idInfo, $e_idClose, $e_idRestart, $e_idBSFolder, $e_idTSadd, $e_idShowResUsage, $e_idLogToFile, $e_idSndVol, $e_idSndApp, $e_idOnClickVolMix, $e_idVolMixMove, $e_idAutoCloseVolMix, $e_idMmsys1, $e_idMmsys2, $e_idMmsys3, $e_exeMsdt
	Local $r = 77, $nCtrl, $i, $n, $a, $hMenu = _GUICtrlMenu_CreatePopup(), $hSubMenu = _GUICtrlMenu_CreateMenu(), $hSubMenuSetMic = _GUICtrlMenu_CreateMenu()
	$a = $g_MyTrayMenuData ; _EnumerateAudioDevices() ; for faster menu generation

	$nCtrl = UBound($a, 2)
	If UBound($a) Then ReDim $a[UBound($a)][UBound($a, 2) + 1]

	_GUICtrlMenu_AddMenuItem($hSubMenu, "About [v" & GetScriptVer() & "]", $e_idSave)
	_GUICtrlMenu_AddMenuItem($hSubMenu, "", 0)
	_GUICtrlMenu_AddMenuItem($hSubMenu, "Exit", $e_idClose)
	_GUICtrlMenu_AddMenuItem($hSubMenu, "", 0)
	_GUICtrlMenu_AddMenuItem($hSubMenu, "Restart [" & StringRight(@ScriptName, 3) & "][" & (@AutoItX64 ? "x64" : "x32") & "][" & (IsAdmin() ? "admin level" : "user level") & "]", $e_idRestart)

	If Not ($exeSndVol = "") Then
		_GUICtrlMenu_AddMenuItem($hSubMenu, "", 0)
		_GUICtrlMenu_AddMenuItem($hSubMenu, "Vol. Mix.: Move it close to tray area (use on Win11 ?)", $e_idVolMixMove)
		If $g_MoveToBottomRightVolMix Then _GUICtrlMenu_CheckMenuItem($hSubMenu, $e_idVolMixMove, True, False)
		_GUICtrlMenu_AddMenuItem($hSubMenu, "OnClick: Vol. Mix.", $e_idOnClickVolMix)
		If $g_OnClickVolMix Then _GUICtrlMenu_CheckMenuItem($hSubMenu, $e_idOnClickVolMix, True, False)
		_GUICtrlMenu_AddMenuItem($hSubMenu, "Auto close Vol. Mix.", $e_idAutoCloseVolMix)
		If $g_AutoCloseVolMix Then _GUICtrlMenu_CheckMenuItem($hSubMenu, $e_idAutoCloseVolMix, True, False)
		If Not $g_OnClickVolMix Then _GUICtrlMenu_SetItemGrayed($hSubMenu, $e_idAutoCloseVolMix, True, False)
		_GUICtrlMenu_AddMenuItem($hSubMenu, "", 0)
		_GUICtrlMenu_AddMenuItem($hSubMenu, "Open Volume Mixer", $e_idSndVol)
		_GUICtrlMenu_AddMenuItem($hSubMenu, "Open App volume", $e_idSndApp) ; ShellExecute("ms-settings:apps-volume") ; new on 0.2022.4.15
		_GUICtrlMenu_AddMenuItem($hSubMenu, "", 0)
		_GUICtrlMenu_AddMenuItem($hSubMenu, "Playback devices", $e_idMmsys1)
		_GUICtrlMenu_AddMenuItem($hSubMenu, "Recording devices", $e_idMmsys2)
		_GUICtrlMenu_AddMenuItem($hSubMenu, "Sounds", $e_idMmsys3)
		If Not ($exeMsdt = "") Then _GUICtrlMenu_AddMenuItem($hSubMenu, "Troubleshoot sound problems", $e_exeMsdt)
	EndIf

	_GUICtrlMenu_AddMenuItem($hSubMenu, "", 0)
	_GUICtrlMenu_AddMenuItem($hSubMenu, "Show resources usage (for debug)", $e_idShowResUsage)

	_GUICtrlMenu_AddMenuItem($hSubMenu, "", 0)
	_GUICtrlMenu_AddMenuItem($hSubMenu, "Show enabled Player/Mic. (for debug)", $e_idInfo)

	_GUICtrlMenu_AddMenuItem($hSubMenu, "", 0)

	If $g_ConsoleLogToFile Then
		_GUICtrlMenu_AddMenuItem($hSubMenu, "Log events to file is ON (for debug)", $e_idLogToFile)
		_GUICtrlMenu_CheckMenuItem($hSubMenu, $e_idLogToFile, True, False)
	Else
		_GUICtrlMenu_AddMenuItem($hSubMenu, "Log events to file is OFF (for debug)", $e_idLogToFile)
	EndIf

	_GUICtrlMenu_AddMenuItem($hSubMenu, "", 0)
	If _CanBeAdmin() = 1 Then
		_GUICtrlMenu_AddMenuItem($hSubMenu, "Elevate to Admin (for greater functionality)", $e_idTSadd)
	Else
		If _TaskSched_Installed() Then
			_GUICtrlMenu_AddMenuItem($hSubMenu, "Remove from TaskSched AtLogOn", $e_idTSadd)
		Else
			_GUICtrlMenu_AddMenuItem($hSubMenu, "Add to TaskSched AtLogOn", $e_idTSadd)
		EndIf
	EndIf


	_GUICtrlMenu_AddMenuItem($hSubMenu, "", 0)
	_GUICtrlMenu_AddMenuItem($hSubMenu, "Browse self folder", $e_idBSFolder)

	_GUICtrlMenu_AddMenuItem($hMenu, "more ..", 0, $hSubMenu)
	_GUICtrlMenu_AddMenuItem($hMenu, "", 0)
	_GUICtrlMenu_AddMenuItem($hMenu, "Set Microphone on ..", 0, $hSubMenuSetMic)

	For $n = 0 To UBound($a) - 1
		$a[$n][$nCtrl] = 1100 + $n
		If $a[$n][0] = "Microphone" Then
			_GUICtrlMenu_AddMenuItem($hSubMenuSetMic, "Set Microphone on: " & $a[$n][3], $a[$n][$nCtrl])
			If Int($a[$n][4]) == 1 Then
				_GUICtrlMenu_CheckMenuItem($hSubMenuSetMic, $a[$n][$nCtrl], True, False)
				_GUICtrlMenu_SetItemGrayed($hSubMenuSetMic, $a[$n][$nCtrl], True, False)
				TraySetToolTip_Mic($a[$n][3])
			EndIf
		EndIf
	Next

	_GUICtrlMenu_AddMenuItem($hMenu, "", 0)
	For $n = 0 To UBound($a) - 1
		$a[$n][$nCtrl] = 1100 + $n
		If $a[$n][0] = "Playback" Then
			_GUICtrlMenu_AddMenuItem($hMenu, "Set Playback on: " & $a[$n][3], $a[$n][$nCtrl])
			If Int($a[$n][4]) == 1 Then
				_GUICtrlMenu_CheckMenuItem($hMenu, $a[$n][$nCtrl], True, False)
				_GUICtrlMenu_SetItemGrayed($hMenu, $a[$n][$nCtrl], True, False)
			EndIf
		EndIf
	Next
	_GUICtrlMenu_AddMenuItem($hMenu, "", 0)
	_GUICtrlMenu_AddMenuItem($hMenu, "Cancel menu", $e_idCancel)
	GUISetState(@SW_SHOW, $hGui_GuiContextMenu) ; < ----- this solves the problem   =)
;~ 	ConsoleWrite(@TAB & TimerDiff($t) & @CRLF)
	$i = _GUICtrlMenu_TrackPopupMenu($hMenu, $hGui_GuiContextMenu, -1, -1, 1, 1, 2)
	ConsoleLog("- TrackPopupMenu = " & $i & @CRLF)
	Switch $i
		Case 0
			; nothing
			ConsoleLog(@TAB & "Menu: abandoned" & @CRLF & @CRLF)

		Case $e_idCancel
			ConsoleLog(@TAB & "Menu: canceled" & @CRLF & @CRLF)

		Case $e_idSndVol
			ConsoleLog(@TAB & "Menu: ShellExecute(" & $exeSndVol & ")" & @CRLF & @CRLF)
			ShellExecute($exeSndVol, "-m " & ($m[1] * 65536) + $m[0])

		Case $e_idSndApp
			ConsoleLog(@TAB & "Menu: ShellExecute(ms-settings:apps-volume)" & @CRLF & @CRLF)
			ShellExecute("ms-settings:apps-volume")

		Case $e_idMmsys1
			ShellExecute("control.exe", "mmsys.cpl,,0")

		Case $e_idMmsys2
			ShellExecute("control.exe", "mmsys.cpl,,1")

		Case $e_idMmsys3
			ShellExecute("control.exe", "mmsys.cpl,,2")

		Case $e_exeMsdt
			ConsoleLog(@TAB & "Menu: ShellExecute(" & $exeMsdt & ")" & @CRLF & @CRLF)
			ShellExecute($exeMsdt, "-id AudioPlaybackDiagnostic")

		Case $e_idTSadd
			If _CanBeAdmin() = 1 Then
				ShellExecute(@AutoItExe, ( @Compiled ? "" : '"' & @ScriptFullPath & '" ' ) & '/Elevated', "", "runas")
			Else
				If _TaskSched_Installed() Then
					_TaskSched_Uninstall()
					ConsoleLog(@TAB & "Menu: TaskSched removed" & @CRLF & @CRLF)
				Else
					_TaskSched_Install()
					ConsoleLog(@TAB & "Menu: TaskSched added" & @CRLF & @CRLF)
				EndIf
				Sleep(1000)
				_TaskSched_Installed(0) ; reload/confirm status
			EndIf

		Case $e_idShowResUsage
			ConsoleLog(@TAB & "Menu: ShowResourcesUsage" & @CRLF & @CRLF)
			ShowResourcesUsage()

		Case $e_idVolMixMove ; $g_MoveToBottomRightVolMix
			$g_MoveToBottomRightVolMix = Int( Not $g_MoveToBottomRightVolMix )
			ConsoleLog(@TAB & "Menu: VolMixMove = " & $g_MoveToBottomRightVolMix & @CRLF & @CRLF, 1)
			IniSet("MoveToBottomRightVolMix", $g_MoveToBottomRightVolMix, Default, ' ; for fixing Win11, forcing users to use their -much disliked by me- "new and improved" interface')

		Case $e_idOnClickVolMix
			If $g_OnClickVolMix Then
				$g_OnClickVolMix = 0
				TraySetOnEvent($TRAY_EVENT_PRIMARYDOWN, "MyTrayMenu")
			Else
				$g_OnClickVolMix = 1
				TraySetOnEvent($TRAY_EVENT_PRIMARYDOWN, "MyTrayMenuVolMix")
			EndIf
			ConsoleLog(@TAB & "Menu: OnClickVolMix = " & $g_OnClickVolMix & @CRLF & @CRLF, 1)
			IniSet("OnClickVolMix", $g_OnClickVolMix)

		Case $e_idAutoCloseVolMix
			If $g_AutoCloseVolMix Then
				$g_AutoCloseVolMix = 0
			Else
				$g_AutoCloseVolMix = 1
			EndIf
			ConsoleLog(@TAB & "Menu: AutoCloseVolMix = " & $g_AutoCloseVolMix & @CRLF & @CRLF, 1)
			IniSet("AutoCloseVolMix", $g_AutoCloseVolMix)

		Case $e_idLogToFile
			If $g_ConsoleLogToFile Then
				$g_ConsoleLogToFile = 0
			Else
				$g_ConsoleLogToFile = 1
			EndIf
			ConsoleLog(@TAB & "Menu: LogToFile = " & $g_ConsoleLogToFile & @CRLF & @CRLF, 1)
			IniSet("LogToFile", $g_ConsoleLogToFile)

		Case $e_idBSFolder
			ConsoleLog(@TAB & "Menu: Browse self folder" & @CRLF & @CRLF)
			ShellExecute(@ScriptDir)

		Case $e_idSave
			ConsoleLog(@TAB & "Menu: About" & @CRLF & @CRLF)
			ShellExecute(@AutoItExe, ( @Compiled ? "" : '"' & @ScriptFullPath & '" ' ) & '/About')

		Case $e_idInfo
			ConsoleLog(@TAB & "Menu: DoEnumerateAudioDevices" & @CRLF & @CRLF)
			ShellExecute(@AutoItExe, ( @Compiled ? "" : '"' & @ScriptFullPath & '" ' ) & "/DoEnumerateAudioDevices")

		Case $e_idClose
			ConsoleLog(@TAB & "Menu: Exit" & @CRLF & @CRLF)
			Exit 0

		Case $e_idRestart
			ConsoleLog(@TAB & "Menu: Restart" & @CRLF & @CRLF)
			AutoItWinSetTitle($g_ExeName & "_RestartMain")
			ShellExecute(@AutoItExe, ( @Compiled ? "" : '"' & @ScriptFullPath & '" ' ))
			Exit 0

		Case Else
			For $n = 0 To UBound($a) - 1
				If $i = $a[$n][$nCtrl] Then
					If $a[$n][0] = "Microphone" Then
						IniSet("UserMicSel", $a[$n][2] & "|" & $a[$n][3])
						$g_UserMicSel[0] = $a[$n][2] ; {0.0.1.00000000}.{c475b847-aafb-4d6f-9752-0bd60036c0f2}
						$g_UserMicSel[1] = $a[$n][3] ; Desktop Microphone (Cinema - Microsoft LifeCam.)
					EndIf
					$r = ShellExecuteWait(@AutoItExe, ( @Compiled ? "" : '"' & @ScriptFullPath & '" ' ) & "/ErrorStdOut " & $a[$n][2], @ScriptDir)
					ConsoleLog(@TAB & 'Menu: Set ' & $a[$n][0] & ' on: "' & $a[$n][3] & '" = ' & ($r = 3 ? "OK" : "FAILED, error level " & $r) & @CRLF & @CRLF)
				EndIf
			Next
			If $r = 77 Then ConsoleLog(@TAB & 'Menu: unknown event # ' & $i & @CRLF & @CRLF)

	EndSwitch
	_GUICtrlMenu_DestroyMenu($hSubMenu)
	_GUICtrlMenu_DestroyMenu($hMenu)
	GUISetState(@SW_HIDE, $hGui_GuiContextMenu)
EndFunc   ;==>MyTrayMenu

Func MyTrayMenu_None()
EndFunc   ;==>MyTrayMenu_None

Func MyTrayMenu_On()
	AdlibUnRegister("MyTrayMenu_On")
	TraySetOnEvent($TRAY_EVENT_SECONDARYUP, "MyTrayMenu")
EndFunc   ;==>MyTrayMenu_On

Func EventListener_Init()

	Local $iAccess = StringInStr(@OSArch, "64") ? BitOR($KEY_NOTIFY, $KEY_WOW64_64KEY) : $KEY_NOTIFY
	If @AutoItX64 Then $iAccess = $KEY_NOTIFY
	Local $sRegAudio = "SOFTWARE\Microsoft\Windows\CurrentVersion\MMDevices\Audio"
	$g_EventListener_hKey = _WinAPI_RegOpenKey($HKEY_LOCAL_MACHINE, $sRegAudio, $iAccess)
	If @error Or Not $g_EventListener_hKey Then
		Return SetError(4, @extended, 4)
		;MsgBox(BitOR($MB_ICONERROR, $MB_SYSTEMMODAL), @extended, _WinAPI_GetErrorMessage(@extended), 30)
	EndIf

	OnAutoItExitRegister("EventListener_OnAutoIt_Exit")

	$g_EventListener_tEvents = DllStructCreate("handle Event[1];")
	$g_EventListener_tEvents.Event(1) = _WinAPI_CreateEvent(0, True, False)

	If Not _WinAPI_RegNotifyChangeKeyValue($g_EventListener_hKey, BitOR($REG_NOTIFY_CHANGE_NAME, $REG_NOTIFY_CHANGE_LAST_SET, $REG_NOTIFY_CHANGE_ATTRIBUTES), 1, 1, $g_EventListener_tEvents.Event(1)) Then
		Return SetError(5, @extended, 5)
	EndIf
	Return 0
EndFunc   ;==>EventListener_Init

Func IfMiddleClick_RunAppVolume()
	Local $aMosePos = MouseGetPos()
	If @error Then Return
	If $aMosePos[0] < $g_EVENT_MOUSEOVER_aMousePos[0] + 5 And _
		$aMosePos[0] > $g_EVENT_MOUSEOVER_aMousePos[0] - 5 And _
		$aMosePos[1] < $g_EVENT_MOUSEOVER_aMousePos[1] + 5 And _
		$aMosePos[1] > $g_EVENT_MOUSEOVER_aMousePos[1] - 5 Then ; I guess we are in the area of the tray icon.
		$g_IfMiddleClick_RunAppVolume = TimerInit()
		ShellExecute("ms-settings:apps-volume")
	EndIf
	While _IsPressed("04")
		Sleep(50)
	WEnd
EndFunc

Func EventListener_Loop()
	Local $iEvent, $i
	While 1

		If TimerDiff($g_EVENT_MOUSEOVER_hTimer) < 5000 And _IsPressed("04") Then IfMiddleClick_RunAppVolume()

		$iEvent = _WinAPI_WaitForMultipleObjects(1, $g_EventListener_tEvents, False, 100)
		Switch $iEvent
			Case 0
				AdlibRegister("AudioDevices_SetMatchingMic")
				If Not _WinAPI_RegNotifyChangeKeyValue($g_EventListener_hKey, BitOR($REG_NOTIFY_CHANGE_NAME, $REG_NOTIFY_CHANGE_LAST_SET, $REG_NOTIFY_CHANGE_ATTRIBUTES), 1, 1, $g_EventListener_tEvents.Event(1)) Then
					$i = @extended
					$sErrorHandler &= "RegNotifyChangeKeyValue ERROR: " & _WinAPI_GetErrorMessage($i) & @CRLF
					Return SetError(6, $i, 6)
				EndIf
			Case -1 ;Error
				$sErrorHandler &= "WaitForMultipleObjects ERROR: " & _WinAPI_GetLastError() & @CRLF
				Return SetError(7, 0, 7)
			Case Else
				Return 0
		EndSwitch
	WEnd
EndFunc   ;==>EventListener_Loop

Func EventListener_OnAutoIt_Exit()
	If $g_EventListener_hKey Then
		_WinAPI_CloseHandle($g_EventListener_tEvents.Event(1))
		If Not @error Then
			$g_EventListener_tEvents = 0
		Else
			$sErrorHandler &= "tEvents CloseHandle FAILED = " & _WinAPI_GetLastError() & @CRLF
		EndIf
		_WinAPI_RegCloseKey($g_EventListener_hKey)
		If Not @error Then
			$g_EventListener_hKey = 0
		Else
			$sErrorHandler &= "hKey RegCloseKey FAILED = " & _WinAPI_GetErrorMessage(@extended) & @CRLF
		EndIf
	EndIf
EndFunc   ;==>EventListener_OnAutoIt_Exit

Func AudioDevices_SetMatchingMic()
	AdlibUnRegister("AudioDevices_SetMatchingMic")
	Local Static $dP[5], $dM[5], $dIco = ""
	Local $dMic = 0, $i = 5, $s = "", $b, $c = 0, $m, $n, $uA, $d, $a = _EnumerateAudioDevices()
	$uA = UBound($a) - 1
;~ 	$uB = UBound($b) -1
	For $n = 0 To $uA
		If Not $a[$n][4] Then ContinueLoop
		Switch $a[$n][0]
			Case "Playback"
				If $dP[2] <> $a[$n][2] Then
					$c += 1
					For $m = 0 To 4
						$dP[$m] = $a[$n][$m]
					Next
				EndIf
				$d = StringSplit($dP[3], " ") ; ..would be nice to have it's own icon
				If FileGetSize(@ScriptDir & "\MicSwitcher." & $d[1] & ".ico") Then
					If Not ($dIco = $d[1]) Then TraySetIcon(@ScriptDir & "\MicSwitcher." & $d[1] & ".ico")
					$dIco = $d[1]
				Else
;~ 					If Not ($dIco = "") Then
					Switch $d[1]
						Case "speakers", "speaker"
							If @Compiled Then
								TraySetIcon(@ScriptFullPath, 201)
							Else
								TraySetIcon(@ScriptDir & "\MicSwitcher.speakers.ico")
							EndIf
							$dIco = $d[1]

						Case "Headset", "Headphones"
							If @Compiled Then
								TraySetIcon(@ScriptFullPath, 202)
							Else
								TraySetIcon(@ScriptDir & "\MicSwitcher.Headset.ico")
							EndIf
							$dIco = $d[1]

						Case Else
							If @Compiled Then
								TraySetIcon(@ScriptFullPath)
							Else
								TraySetIcon(@ScriptDir & "\MicSwitcher.ico")
							EndIf

					EndSwitch
;~ 					EndIf
					$dIco = ""
				EndIf

			Case "Microphone"
				If $dM[2] <> $a[$n][2] Then
					$c += 2
					For $m = 0 To 4
						$dM[$m] = $a[$n][$m]
					Next
				EndIf

		EndSwitch
	Next

	If BitAND($c, 2) Then TraySetToolTip_Mic($dM[3])

	If BitAND($c, 1) Then
		$b = StringSplit($dP[3], "()", 0)
		If UBound($b) > 2 Then $s = $b[2]
		ConsoleLog('found "' & $s & '"' & @CRLF)
		If $s = "" Then
			ConsoleLog(@TAB & $s & ' = not found ?, just 1 device ?, remote audio ?' & @CRLF & @CRLF)
			Return $a
		Else
			For $n = 0 To UBound($a) - 1
				If $a[$n][0] = "Microphone" Then
					If Int($a[$n][4]) == 1 Then $dMic = $n
					If StringInStr($a[$n][3], $s) Then             ;                                              ObjCreateInterface: "Warning: This feature is experimental."
;~ 									$i = _SwitchToDevice($a[$n][2])
						$i = ShellExecuteWait(@AutoItExe, ( @Compiled ? "" : '"' & @ScriptFullPath & '" ' ) & "/ErrorStdOut " & $a[$n][2], @ScriptDir)
						ConsoleLog(@TAB & $s & ": apply = " & $i & @CRLF)             ; this crashes the script. Run separately.
						Switch $i
							Case 3
								ConsoleLog(@TAB & $s & ": Mic. is ON = True" & @CRLF & @CRLF)
								TraySetToolTip_Mic($a[$n][3])
							Case 4
								ConsoleLog(@TAB & $s & ": Mic. is ON = False" & @CRLF & @CRLF)
							Case 1
								ConsoleLog(@TAB & $s & ": Mic. is ON = Crashed" & @CRLF & @CRLF)
							Case Else
								ConsoleLog(@TAB & $s & ': Mic. is "no clue !"' & @CRLF & @CRLF)
						EndSwitch
						ExitLoop
					EndIf
				Else
					ContinueLoop
				EndIf
			Next
		EndIf
		If $i = 5 Then
			If StringLen($g_UserMicSel[0]) And StringLen($g_UserMicSel[1]) Then
				For $n = 0 To UBound($a) - 1
					If $g_UserMicSel[0] = $a[$n][2] Then
						$i = ShellExecuteWait(@AutoItExe, ( @Compiled ? "" : '"' & @ScriptFullPath & '" ' ) & "/ErrorStdOut " & $a[$n][2], @ScriptDir)
						ConsoleLog(@TAB & $s & ": apply = " & $i & " (user default)" & @CRLF)             ; this crashes the script. Run separately.
						Switch $i
							Case 3
								ConsoleLog(@TAB & $s & ": Mic. is ON = True" & @CRLF & @CRLF)
								TraySetToolTip_Mic($a[$n][3])
							Case 4
								ConsoleLog(@TAB & $s & ": Mic. is ON = False" & @CRLF & @CRLF)
							Case 1
								ConsoleLog(@TAB & $s & ": Mic. is ON = Crashed" & @CRLF & @CRLF)
							Case Else
								ConsoleLog(@TAB & $s & ': Mic. is "no clue !"' & @CRLF & @CRLF)
						EndSwitch
						ExitLoop
					EndIf
				Next
				If $i = 5 Then ConsoleLog(@TAB & $s & ' = "default mic" not found. Using "' & $a[$dMic][3] & '"' & @CRLF & @CRLF)
			Else
				ConsoleLog(@TAB & $s & ' = no "default mic" ever selected. Using "' & $a[$dMic][3] & '"' & @CRLF & @CRLF)
			EndIf
		EndIf
	EndIf
	Return $a
EndFunc   ;==>AudioDevices_SetMatchingMic

Func TraySetToolTip_Mic($sStr = Default)
	Local Static $sMicName = "" ; "No Mic. in use" & @CR & @CR & @CR & " " ; "Mic. in use is:" & @CR & $dM[3] & @CR & @CR & " "
	If $sStr = Default Or $sMicName = $sStr Then Return $sMicName
	$sMicName = $sStr
	TraySetToolTip(($sMicName = "" ? "No Mic. in use" : "Mic. in use is:") & @CR & $sMicName & @CR & @CR & " ")
	Return $sStr
EndFunc   ;==>TraySetToolTip_Mic

#EndRegion tray

Func main()

	AutoItWinSetTitle($g_ExeName & "_Main")
	$hGui_Main = WinGetHandle($g_ExeName & "_Main")
	$hGui_GuiContextMenu = GUICreate($g_ExeName & "_GuiContextMenu", 0, 0, -1, -1, BitOR($WS_SYSMENU, $WS_POPUP), $WS_EX_TOPMOST, $hGui_Main)
	GUISetState(@SW_HIDE, $hGui_GuiContextMenu)
	ConsoleLog("-= Started (v" & GetScriptVer() & ") =-" & @CRLF)
	MemLeakTooMuch()
	Opt("TrayMenuMode", 3)     ; The default tray menu items will not be shown and items are not checked when selected. These are options 1 and 2 for TrayMenuMode.
	Opt("TrayOnEventMode", 1)     ; Enable TrayOnEventMode.

;~ 	TraySetOnEvent($TRAY_EVENT_PRIMARYDOUBLE, "RunAboutCmd")
	TraySetOnEvent($TRAY_EVENT_PRIMARYDOUBLE, "MyTrayMenu")
	If $g_OnClickVolMix Then
		TraySetOnEvent($TRAY_EVENT_PRIMARYDOWN, "MyTrayMenuVolMix")
	Else
		TraySetOnEvent($TRAY_EVENT_PRIMARYDOWN, "MyTrayMenu")
	EndIf

	TraySetOnEvent($TRAY_EVENT_SECONDARYUP, "MyTrayMenu")
	TraySetOnEvent($TRAY_EVENT_MOUSEOVER, "MyTrayMenuReqData")

	TraySetState($TRAY_ICONSTATE_SHOW)     ; Show the tray menu.
	TraySetState($TRAY_ICONSTATE_STOPFLASH)     ; ..if it was flashing
	TraySetToolTip_Mic("")

	AudioDevices_SetMatchingMic()
	EventListener_Init()
	If @error Or $iErrorHandler Then
		TraySetState($TRAY_ICONSTATE_HIDE)     ; Show the tray menu.
		About_Box(1)
		MsgBox(262144 + 16, $g_ExeName, "This, unfortunately, will not run in your system.", 30)
		Exit 8
	EndIf

	AdlibRegister("OnMemLeakRestart", 3600000)     ; once an hour

	_TaskSched_Installed() ; ..pre load state

	If _CanBeAdmin() = 1 And _TaskSched_Installed() Then Run('schtasks.exe /Run /TN "\' & $g_ExeName & '"', '', @SW_HIDE)

	While 1
		EventListener_Loop()
		If @error Then Exit 15
	WEnd
EndFunc   ;==>main

Func _DoEnumerateAudioDevices()
	Local $a = _EnumerateAudioDevices()
	_DebugArrayDisplay($a, $g_ExeName & " :  _EnumerateAudioDevices()  (for Debug)", Default, Default, @TAB & "|", "Type|GUID|ID|Name|Selected")
EndFunc   ;==>_DoEnumerateAudioDevices

#Region About

Func About_Box($1 = 0)
	; https://www.autoitscript.com/forum/topic/44889-create-an-about-window/?do=findComment&comment=334569
	Opt("GUICloseOnESC", 1)
	Opt("GUIOnEventMode", 1)
	AutoItWinSetTitle($g_ExeName & "_About|" & @AutoItPID)

	Local $aGui[10]
	$aGui[0] = 250
	$aGui[1] = 150
	If $1 Then $aGui[1] += 100     ; "This, unfortunately, will not run in your system."
	$aGui[2] = -1
	$aGui[3] = -1
	$aGui[4] = "About:  " & $g_ExeName
	If $1 Then $aGui[4] = "Error:  " & $g_ExeName
	If WinExists("[TITLE:" & $aGui[4] & ";CLASS:AutoIt v3 GUI;]") Then Exit 0
	$aGui[5] = GUICreate($aGui[4], $aGui[0], $aGui[1], $aGui[2], $aGui[3], BitOR($WS_CAPTION, $WS_SYSMENU), $WS_EX_TOPMOST)
	$aGui[6] = $g_ExeName
	GUISetOnEvent($GUI_EVENT_CLOSE, "About_OK")

	If @Compiled Then
		GUICtrlCreateIcon($aGui[6] & ".exe", -1, 11, 11)
	Else
		GUICtrlCreateIcon(@ScriptDir & "\MicSwitcher.ico", -1, 11, 11)
		GUISetIcon(@ScriptDir & "\MicSwitcher.ico")
	EndIf

	GUICtrlCreateLabel($aGui[6], 60, 15, $aGui[0], 20)
	GUICtrlSetFont(-1, 10, 800, 0, "Arial")     ; bold
	GUICtrlCreateLabel("v. " & GetScriptVer() & "" & @CR & "by argumentum", 60, 40, $aGui[0] - 65, 40)

	Local $www = GUICtrlCreateLabel("www.autoitscript.com/forum/", 59, 75, 140, 15)
	GUICtrlSetFont($www, 8.5, -1, 4)     ; underlined
	GUICtrlSetColor($www, _WinAPI_SwitchColor(_WinAPI_GetSysColor($COLOR_HOTLIGHT)))
	GUICtrlSetCursor($www, 0)
	GUICtrlSetOnEvent(-1, "About_WWW")

	GUICtrlCreateButton("OK", ($aGui[0] / 2) - 50, $aGui[1] - 40, 100, 25, BitOR($GUI_SS_DEFAULT_BUTTON, $BS_DEFPUSHBUTTON))
	GUICtrlSetState(-1, $GUI_FOCUS)
	GUICtrlSetOnEvent(-1, "About_OK")

	If $1 Then
		GUICtrlCreateLabel("This, unfortunately, failed to run on your system", 59, 120, 140, 65)
		GUICtrlSetFont(-1, 10, 800, 0, "Arial")     ; bold
	EndIf

	GUISetState(@SW_SHOW, $aGui[5])

	While 1
		Sleep(100)
	WEnd
EndFunc   ;==>About_Box

Func About_WWW()
	ShellExecute("https://www.autoitscript.com/forum/files/file/476-setmicforplayback/")
	About_OK()
EndFunc   ;==>About_WWW

Func About_OK()
	GUIDelete()
	If Int(Eval("iErrorHandler")) Then Exit 8
	Exit 0
EndFunc   ;==>About_OK

#EndRegion About

#Region MemLeak

Func OnMemLeakRestart()     ;
	Local Static $iTriggered = 0
	If MemLeakTooMuch() Then
		If Not $iTriggered Then
			$iTriggered = 1
			AdlibRegister("OnMemLeakRestart", 60000)     ; since it triggered, check more often
		EndIf
		If _WinAPI_GetIdleTime() > 600000 Then     ; over 10 min. maybe ?, to not bother the user ?

			ShellExecute(@AutoItExe, ( @Compiled ? "" : '"' & @ScriptFullPath & '" ' ))
			; restart itself if it must, but if you do, keep that in mind while coding.

			Exit 3     ; use the exit code you please other than 1 or 2,
		EndIf     ;                     as AutoIt may use it for itself.
	EndIf
EndFunc   ;==>OnMemLeakRestart

Func MemLeakLog()
	Local $aMethod[5] = ["EXITCLOSE_NORMAL", "EXITCLOSE_BYEXIT", "EXITCLOSE_BYCLICK", "EXITCLOSE_BYLOGOFF", "EXITCLOSE_BYSHUTDOWN"]
	Local $iMethod = Execute("@exitMethod")
	Local $s = @YEAR & "." & @MON & "." & @MDAY & "_" & @HOUR & ":" & @MIN & ":" & @SEC
	$s &= "  -  RunTime: " & _Convert(gTimer()) & "  -  PID: " & @AutoItPID
	If IsInt($iMethod) Then
		$s &= "  -  ExitCode: " & @exitCode
		$s &= "  -  ExitMethod: " & $iMethod & " (" & $aMethod[$iMethod] & ")"
	EndIf
	Local $a = MemLeakUsage()
	ConsoleLog($s & @CRLF & _SQLite_Display2DResult($a, 0, True) & Eval("sErrorHandler") & @CRLF)
	; I save the data from ObjEvent("AutoIt.Error", "_ErrFunc") to $sErrorHandler
EndFunc   ;==>MemLeakLog

Func MemLeakUsage()
	Local $aDesc[14] = ["number of page faults", "peak working set size, in bytes", "current working set size, in bytes", _
			"peak paged pool usage, in bytes", "current paged pool usage, in bytes", "peak nonpaged pool usage, in bytes", _
			"current nonpaged pool usage, in bytes", "current space allocated for the pagefile, in bytes", _
			"peak space allocated for the pagefile, in bytes", "current amount of memory that cannot be shared with other processes, in bytes", _
			"ProcessHandleCount", "ProcessThreadsCount", "GDI objects", "USER objects"]
	Local $a[4][15]
	$a[1][0] = "Starting:"
	$a[2][0] = " Exiting:"
	$a[3][0] = "   Diff.:"
	MemLeakTooMuch()
	Local $b = MemLeakTooMuch(-2), $c = MemLeakTooMuch(-3)
	For $n = 0 To UBound($c) - 1
		$a[0][$n + 1] = $aDesc[$n]
		$a[1][$n + 1] = $b[$n]
		$a[2][$n + 1] = $c[$n]
		$a[3][$n + 1] = $c[$n] - $b[$n]
		Switch $n
			Case 0, 10, 11, 12, 13, 14
				; nothing, is a count number
			Case Else
				$a[3][$n + 1] = ByteSuffix($a[3][$n + 1])
		EndSwitch
		$a[3][$n + 1] &= "  ( x" & (($b[$n] > 0 And $b[$n] > 0) ? Round($c[$n] / $b[$n], 1) : "1") & " )"
	Next
	$a[0][0] = _Convert(gTimer())
	Return $a
EndFunc   ;==>MemLeakUsage

Func MemLeakTooMuch($iPid = 0, $iMutliplier = 4)     ; in case this is for some other PID, but only one  =/
	Local Static $aInitProcessMemoryInfo = 99, $aLastProcessMemoryInfo
	Local $ret = False, $a, $n, $aTemp
	If $iPid = -2 Then Return $aInitProcessMemoryInfo     ; returns the initial readings
	If $iPid = -3 Then Return $aLastProcessMemoryInfo     ; returns the last readings
	If $aInitProcessMemoryInfo = 99 Then
		$aTemp = _WinAPI_GetProcessMemoryInfo($iPid)
		If Not (UBound($aTemp) = 10) Then Return SetError(1, 0, $ret)
		$aLastProcessMemoryInfo = $aTemp
	Else
		$aTemp = _WinAPI_GetProcessMemoryInfo($iPid)
		If Not (UBound($aTemp) = 10) Then Return SetError(1, 0, $ret)
		$aLastProcessMemoryInfo = $aTemp
		For $n = 1 To UBound($aTemp) - 1     ; I don't think "page faults" is leakage, so it's omited.
			If $aTemp[$n] > $aInitProcessMemoryInfo[$n] * $iMutliplier Then $ret = True
		Next
	EndIf
	ReDim $aLastProcessMemoryInfo[14]     ; to hold the next values
	$aLastProcessMemoryInfo[10] = _WinAPI_GetProcessHandleCount($iPid)
	If $aLastProcessMemoryInfo[10] > 2000 Then $ret = True     ; dropping handles ?
	$a = _WinAPI_EnumProcessThreads($iPid)
	$aLastProcessMemoryInfo[11] = UBound($a) - 1
	If $aLastProcessMemoryInfo[11] > 200 Then $ret = True     ; dropping threads ?
	If Not $iPid Then $iPid = -1
	$aLastProcessMemoryInfo[12] = _WinAPI_GetGuiResources(0, $iPid)     ; count of GDI objects.
	$aLastProcessMemoryInfo[13] = _WinAPI_GetGuiResources(1, $iPid)     ; count of USER objects
	If $aLastProcessMemoryInfo[12] > 1000 Then $ret = True     ; dropping GDI objects ?
	If $aLastProcessMemoryInfo[13] > 2000 Then $ret = True     ; dropping USER objects ?
	If $aInitProcessMemoryInfo = 99 Then
		gTimer()     ; init. the timer now, if was not at the top of the script
		$aInitProcessMemoryInfo = $aLastProcessMemoryInfo
		OnAutoItExitRegister("MemLeakLog")     ; ..if you wanna keep a log
	EndIf
	Return $ret
EndFunc   ;==>MemLeakTooMuch

Func _Convert($ms)     ; https://www.autoitscript.com/forum/topic/163621-convert-ms-to-dayhourminsec/?do=findComment&comment=1192334
	Local $day, $hour, $min, $sec
	_TicksToTime($ms, $hour, $min, $sec)
	If $hour > 24 Then
		$day = $hour / 24
		$hour = Mod($hour, 24)
	EndIf
	Return StringReplace(StringFormat("%03i %02i:%02i:%02i", $day, $hour, $min, $sec), "000 ", "")
EndFunc   ;==>_Convert

Func gTimer()
	Local Static $Timer = TimerInit()
	Return TimerDiff($Timer)
EndFunc   ;==>gTimer

Func ByteSuffix($iBytes)     ; https://www.autoitscript.com/autoit3/docs/functions/FileGetSize.htm
	Local $iIndex = 0, $aArray = [' bytes', ' KB', ' MB', ' GB', ' TB', ' PB', ' EB', ' ZB', ' YB']
	While $iBytes > 1023
		$iIndex += 1
		$iBytes /= 1024
	WEnd
	Return Round($iBytes) & $aArray[$iIndex]
EndFunc   ;==>ByteSuffix

#EndRegion MemLeak

#Region misc.

Func __ArrayFromString($s, $sDelim_Col = "|", $sDelim_Row = @CRLF)     ; https://www.autoitscript.com/forum/topic/197277-_arrayfromstring/
	Local $m, $n, $b, $c = StringSplit($s, $sDelim_Row, 3)
	$b = StringSplit($c[0], $sDelim_Col, 3)
	If UBound($c) = 1 Then
		For $m = 0 To UBound($b) - 1
			If $b[$m] == Int($b[$m]) Then $b[$m] = Int($b[$m])
		Next
		Return $b
	EndIf
	Local $a[UBound($c)][UBound($b)]
	For $n = 0 To UBound($c) - 1
		$b = StringSplit($c[$n], $sDelim_Col, 3)
		If UBound($b) > UBound($a, 2) Then Return SetError(1)
		For $m = 0 To UBound($b) - 1
			If $b[$m] == Int($b[$m]) Then $b[$m] = Int($b[$m])
			$a[$n][$m] = $b[$m]
		Next
	Next
	Return $a
EndFunc   ;==>__ArrayFromString

Func _Array_Rotate2D(ByRef $a)
	Local $u2, $n, $m, $u1 = UBound($a, 1) - 1
	If @error Then Return SetError(1)
	$u2 = UBound($a, 2) - 1
	If @error Then
		$u2 = 1
		Local $b[$u2][$u1 + 1]
		For $n = 0 To $u1
			For $m = 0 To 0
				$b[$m][$n] = $a[$n]
			Next
		Next
	Else
		Local $b[$u2 + 1][$u1 + 1]
		For $n = 0 To $u1
			For $m = 0 To $u2
				$b[$m][$n] = $a[$n][$m]
			Next
		Next
	EndIf
	$a = $b
EndFunc   ;==>_Array_Rotate2D

Func IniSet($sKey, $vValue, $sSection = Default, $sAddComment = Default)
	Local Static $ini = IsDeclared("g_iniFile") ? Eval("g_iniFile") : StringTrimRight(@ScriptFullPath, 4) & ".ini"
	If $sSection = Default Then $sSection = "Settings"
	If $sAddComment = Default Then $sAddComment = ""
	If IsInt($vValue) Then
		$vValue = "i:" & $vValue & $sAddComment
	ElseIf IsArray($vValue) Then
		$vValue = "a:" & StringToBinary(_ArrayToString($vValue, Chr(1), -1, -1, Chr(2)))
	ElseIf IsBinary($vValue) Then
		$vValue = "d:" & $vValue
	Else
		$vValue = "s:" & $vValue
	EndIf
	$vValue = IniWrite($ini, $sSection, $sKey, $vValue)
	Return SetError(@error, @extended, $vValue)
EndFunc   ;==>IniSet

Func IniGet($sKey, $vDefault = Default, $sSection = Default, $sAddComment = Default)
	Local Static $ini = IsDeclared("g_iniFile") ? Eval("g_iniFile") : StringTrimRight(@ScriptFullPath, 4) & ".ini"
	If $sSection = Default Then $sSection = "Settings"
	Local $v, $s = IniRead($ini, $sSection, $sKey, Chr(1))
	If $s = Chr(1) Then
		If $vDefault == Default Then
			Return SetError(1, 0, "")
		Else
			IniSet($sKey, $vDefault, $sSection, $sAddComment)
			Return $vDefault
		EndIf
	EndIf

	$v = StringLeft($s, 1)
	Switch $v
		Case "i"
			Return Int(StringTrimLeft($s, 2))
		Case "a"
			Return __ArrayFromString(BinaryToString(StringTrimLeft($s, 2)), Chr(1), Chr(2))
		Case "d"
			Return Binary(StringTrimLeft($s, 2))
		Case Else
			Return String(StringTrimLeft($s, 2))
	EndSwitch
EndFunc   ;==>IniGet

Func ConsoleLog($s, $i = 0)
	Local Static $v = "", $t = 1, $f = "  ============  ", $iResizeLogFile = 1, $LogFile = StringTrimRight(@ScriptFullPath, 4) & ".log", $iPid = @AutoItPID
	If $iResizeLogFile Then
		$iResizeLogFile = 0
		If FileGetSize($LogFile) > 50 * 1024 Then
			If FileExists(StringTrimRight(@ScriptFullPath, 4) & ".prior.log") Then FileDelete(StringTrimRight(@ScriptFullPath, 4) & ".prior.log")
			If FileExists($LogFile) Then FileMove($LogFile, StringTrimRight(@ScriptFullPath, 4) & ".prior.log")
		EndIf
	EndIf
	If TimerDiff($t) > 2000 Then
		$s = @CRLF & $f & @YEAR & "." & @MON & "." & @MDAY & "." & @HOUR & ":" & @MIN & ":" & @SEC & "  [ " & _Convert(gTimer()) & " ] " & $iPid & $f & @CRLF & $s
		$v = ""
		If Not $g_ConsoleLogToFile Then $v = $s
	EndIf
	$t = TimerInit()
	If Not @Compiled Then ConsoleWrite($s)
	If $g_ConsoleLogToFile Or $i Then FileWrite($LogFile, $v & $s)
EndFunc   ;==>ConsoleLog

Func GetScriptVer()
	Local Static $sRet = GetScriptVer_wrkr()
	Return $sRet
EndFunc

Func GetScriptVer_wrkr()
	If @Compiled Then Return FileGetVersion(@AutoItExe) ; same as @ScriptFullPath
	Local $n, $aTmp, $aArray = StringSplit(FileRead(@ScriptFullPath, 2048), @CRLF, 1)
	For $n = 1 To $aArray[0]
		If StringInStr($aArray[$n], "#AutoIt3Wrapper_Res_Fileversion=") = 1 Then
			$aTmp = StringSplit($aArray[$n], "= ;", 0)
			Return $aTmp[2]
		EndIf
	Next
	Return "0.0.0.1"
EndFunc

#EndRegion misc.

#Region _CanBeAdmin

Func _CanBeAdmin()
	Local Static $iCanBeAdmin = ""
	If $iCanBeAdmin == "" Then
		If IsAdmin() Then
			$iCanBeAdmin = 2
			Return $iCanBeAdmin
		EndIf
		Local $i = _IsUACAdmin()
		If $i = 0 And @extended = 1 Then
			$iCanBeAdmin = 1
			Return $iCanBeAdmin
		Else
			$iCanBeAdmin = 0
			Return $iCanBeAdmin
		EndIf
	EndIf
	Return $iCanBeAdmin
EndFunc   ;==>_CanBeAdmin

Func _IsUACAdmin()
	; #FUNCTION# ====================================================================================================================
	; Name ..........: _IsUACAdmin
	; Description ...: Determines if process has Admin privileges and whether running under UAC.
	; Syntax ........: _IsUACAdmin()
	; Parameters ....: None
	; Return values .: Success          - 1 - User has full Admin rights (Elevated Admin w/ UAC)
	;                  Failure          - 0 - User is not an Admin, sets @extended:
	;                                   | 0 - User cannot elevate
	;                                   | 1 - User can elevate
	; Author ........: Erik Pilsits
	; Modified ......:
	; Remarks .......: THE GOOD STUFF: returns 0 w/ @extended = 1 > UAC Protected Admin
	; Related .......:
	; Link ..........:
	; Example .......: No
	; ===============================================================================================================================
	; check elevation
	If StringRegExp(@OSVersion, "_(XP|20(0|3))") Or (Not _IsUACEnabled()) Then     ; XP, XPe, 2000, 2003 > no UAC
		; no UAC available or turned off
		If IsAdmin() Then
			Return SetExtended(0, 1)
		Else
			Return SetExtended(0, 0)
		EndIf
	Else
		; check UAC elevation
		;
		; get process token groups information
		Local $hToken = _Security__OpenProcessToken(_WinAPI_GetCurrentProcess(), $TOKEN_QUERY)
		Local $tTI = _Security__GetTokenInformation($hToken, $TOKENGROUPS)
		_WinAPI_CloseHandle($hToken)
		;
		Local $pTI = DllStructGetPtr($tTI)
		Local $cbSIDATTR = DllStructGetSize(DllStructCreate("ptr;dword"))
		Local $Count = DllStructGetData(DllStructCreate("dword", $pTI), 1)
		Local $pGROUP1 = DllStructGetPtr(DllStructCreate("dword;STRUCT;ptr;dword;ENDSTRUCT", $pTI), 2)
		Local $tGROUP     ;, $sGROUP = ""
		;
		; S-1-5-32-544 > BUILTINAdministrators > $SID_ADMINISTRATORS
		; S-1-16-8192  > Mandatory LabelMedium Mandatory Level (Protected Admin) > $SID_MEDIUM_MANDATORY_LEVEL
		; S-1-16-12288 > Mandatory LabelHigh Mandatory Level (Elevated Admin) > $SID_HIGH_MANDATORY_LEVEL
		; SE_GROUP_USE_FOR_DENY_ONLY = 0x10
		;
		; check SIDs
		Local $inAdminGrp = False, $denyAdmin = False, $elevatedAdmin = False, $sSID
		For $i = 0 To $Count - 1
			$tGROUP = DllStructCreate("ptr;dword", $pGROUP1 + ($cbSIDATTR * $i))
			$sSID = _Security__SidToStringSid(DllStructGetData($tGROUP, 1))
			If StringInStr($sSID, "S-1-5-32-544") Then
				; member of Administrators group
				$inAdminGrp = True
				; check for deny attribute
				If (BitAND(DllStructGetData($tGROUP, 2), 0x10) = 0x10) Then $denyAdmin = True
			ElseIf StringInStr($sSID, "S-1-16-12288") Then
				$elevatedAdmin = True
			EndIf
		Next
		;
		If $inAdminGrp Then
			; check elevated
			If $elevatedAdmin Then
				; check deny status
				If $denyAdmin Then
					; protected Admin CANNOT elevate
					Return SetExtended(0, 0)
				Else
					; elevated Admin
					Return SetExtended(1, 1)
				EndIf
			Else
				; protected Admin
				Return SetExtended(1, 0)
			EndIf
		Else
			; not an Admin
			Return SetExtended(0, 0)
		EndIf
	EndIf
EndFunc   ;==>_IsUACAdmin

Func _IsUACEnabled()
	Return (RegRead("HKLM\SOFTWARE\Microsoft\Windows\CurrentVersion\Policies\System", "EnableLUA") = 1)
EndFunc   ;==>_IsUACEnabled

#EndRegion _CanBeAdmin

#Region TaskSched

Func _TaskSched_BuildXml()
	Local $xml = '<?xml version="1.0" encoding="UTF-16"?>'
	$xml &= @CRLF & '<Task version="1.2" xmlns="http://schemas.microsoft.com/windows/2004/02/mit/task">'
	$xml &= @CRLF & '  <RegistrationInfo>'
	$xml &= @CRLF & '    <Date>' & @YEAR & '-' & @MON & '-' & @MDAY & 'T00:00:00</Date>'
	$xml &= @CRLF & '    <Author>' & @ComputerName & '\' & @UserName & '</Author>'
	$xml &= @CRLF & '    <Description>run "' & @ScriptName & '" at log on</Description>'
	$xml &= @CRLF & '    <URI>\' & $g_ExeName & '</URI>'
	$xml &= @CRLF & '  </RegistrationInfo>'
	$xml &= @CRLF & '  <Triggers>'
	$xml &= @CRLF & '    <LogonTrigger>'
	$xml &= @CRLF & '      <Enabled>true</Enabled>'
	$xml &= @CRLF & '      <UserId>' & @ComputerName & '\' & @UserName & '</UserId>'
	$xml &= @CRLF & '      <Delay>PT10S</Delay>'
	$xml &= @CRLF & '    </LogonTrigger>'
	$xml &= @CRLF & '  </Triggers>'
	$xml &= @CRLF & '  <Principals>'
	$xml &= @CRLF & '    <Principal id="Author">'
	$xml &= @CRLF & '      <UserId>' & @ComputerName & '\' & @UserName & '</UserId>'
	$xml &= @CRLF & '      <LogonType>InteractiveToken</LogonType>'
	If IsAdmin() Then
		$xml &= @CRLF & '      <RunLevel>HighestAvailable</RunLevel>'
	Else
		$xml &= @CRLF & '      <RunLevel>LeastPrivilege</RunLevel>'
	EndIf
	$xml &= @CRLF & '    </Principal>'
	$xml &= @CRLF & '  </Principals>'
	$xml &= @CRLF & '  <Settings>'
	$xml &= @CRLF & '    <MultipleInstancesPolicy>IgnoreNew</MultipleInstancesPolicy>'
	$xml &= @CRLF & '    <DisallowStartIfOnBatteries>false</DisallowStartIfOnBatteries>'
	$xml &= @CRLF & '    <StopIfGoingOnBatteries>false</StopIfGoingOnBatteries>'
	$xml &= @CRLF & '    <AllowHardTerminate>true</AllowHardTerminate>'
	$xml &= @CRLF & '    <StartWhenAvailable>false</StartWhenAvailable>'
	$xml &= @CRLF & '    <RunOnlyIfNetworkAvailable>false</RunOnlyIfNetworkAvailable>'
	$xml &= @CRLF & '    <IdleSettings>'
	$xml &= @CRLF & '      <StopOnIdleEnd>false</StopOnIdleEnd>'
	$xml &= @CRLF & '      <RestartOnIdle>false</RestartOnIdle>'
	$xml &= @CRLF & '    </IdleSettings>'
	$xml &= @CRLF & '    <AllowStartOnDemand>true</AllowStartOnDemand>'
	$xml &= @CRLF & '    <Enabled>true</Enabled>'
	$xml &= @CRLF & '    <Hidden>false</Hidden>'
	$xml &= @CRLF & '    <RunOnlyIfIdle>false</RunOnlyIfIdle>'
	$xml &= @CRLF & '    <WakeToRun>false</WakeToRun>'
	$xml &= @CRLF & '    <ExecutionTimeLimit>PT0S</ExecutionTimeLimit>'
	$xml &= @CRLF & '    <Priority>7</Priority>'
	$xml &= @CRLF & '  </Settings>'
	$xml &= @CRLF & '  <Actions Context="Author">'
	$xml &= @CRLF & '    <Exec>'
	$xml &= @CRLF & '      <Command>' & @AutoItExe & '</Command>'
	$xml &= @CRLF & '      <Arguments> ' & ( @Compiled ? "" : ' "' & @ScriptFullPath & '" ' ) & '/FromTaskSched</Arguments>'
	$xml &= @CRLF & '      <WorkingDirectory>' & @ScriptDir & '</WorkingDirectory>'
	$xml &= @CRLF & '    </Exec>'
	$xml &= @CRLF & '  </Actions>'
	$xml &= @CRLF & '</Task>'
	Return $xml
EndFunc   ;==>_TaskSched_BuildXml

Func _TaskSched_Install()
	Local $sFileXML = @TempDir & '\$~schtsk_' & @MIN & @SEC & @MSEC & '.xml'
	Local $xml = _TaskSched_BuildXml()
	FileDelete($sFileXML)
	FileWrite($sFileXML, $xml)
	If FileRead($sFileXML) <> $xml Then Return False
;~ 	ConsoleWrite('_TaskSched_Install >\' & $g_ExeName & '<' & @CRLF)
	If RunWait('schtasks.exe /Create /XML "' & $sFileXML & '" /TN "\' & $g_ExeName & '"', '', @SW_HIDE) <> 0 Then
		FileDelete($sFileXML)
		Return False
	EndIf
	FileDelete($sFileXML)
	Return True
EndFunc   ;==>_TaskSched_Install

Func _TaskSched_Uninstall()
;~ 	ConsoleWrite('_TaskSched_Uninstall >\' & $g_ExeName & '<' & @CRLF)
	If RunWait('schtasks.exe /Delete /F /TN "\' & $g_ExeName & '"', '', @SW_HIDE) <> 0 Then Return False
	Return True
EndFunc   ;==>_TaskSched_Uninstall

Func _TaskSched_Installed($iDontCheckAgain = 1)
	Local Static $i_Installed = -1
	If $i_Installed > -1 And $iDontCheckAgain Then Return $i_Installed
	If $iDontCheckAgain = 0 Or $i_Installed = -1 Then $i_Installed = Int(_TaskSched_AlreadyInstalled())
	Return $i_Installed
EndFunc   ;==>_TaskSched_Installed
Func _TaskSched_AlreadyInstalled()
	If RunWait('schtasks.exe /Query /TN "\' & $g_ExeName & '"', '', @SW_HIDE) <> 0 Then Return False
	Return True
EndFunc   ;==>_TaskSched_AlreadyInstalled

#EndRegion TaskSched

